import os
import pathlib
import shutil
import traceback
from datetime import datetime
from typing import List, Tuple, Union
from copy import deepcopy
from pathlib import Path

from PySide2 import QtCore, QtWidgets, QtGui
from PySide2.QtWidgets import QListWidgetItem, QMessageBox

from sohstationviewer.conf import constants
from sohstationviewer.model.data_loader import DataLoader
from sohstationviewer.model.data_type_model import DataTypeModel

from sohstationviewer.view.calendar.calendar_dialog import CalendarDialog
from sohstationviewer.view.db_config.channel_dialog import ChannelDialog
from sohstationviewer.view.db_config.data_type_dialog import DataTypeDialog
from sohstationviewer.view.db_config.param_dialog import ParamDialog
from sohstationviewer.view.db_config.plot_type_dialog import PlotTypeDialog
from sohstationviewer.view.file_information.get_file_information import \
    extract_data_set_info
from sohstationviewer.view.file_list_widget import FileListItem
from sohstationviewer.view.plotting.gps_plot.extract_gps_data import \
    extract_gps_data
from sohstationviewer.view.plotting.gps_plot.gps_dialog import GPSDialog
from sohstationviewer.view.plotting.time_power_squared_dialog import (
    TimePowerSquaredDialog
)
from sohstationviewer.view.plotting.waveform_dialog import WaveformDialog
from sohstationviewer.view.search_message.search_message_dialog import (
    SearchMessageDialog
)
from sohstationviewer.view.plotting.state_of_health_widget import SOHWidget
from sohstationviewer.view.help_view import HelpBrowser
from sohstationviewer.view.ui.main_ui import UIMainWindow
from sohstationviewer.view.util.enums import LogType
from sohstationviewer.view.util.functions import (
    check_chan_wildcards_format, check_masspos)
from sohstationviewer.view.channel_prefer_dialog import ChannelPreferDialog

from sohstationviewer.controller.processing import detect_data_type
from sohstationviewer.controller.util import (
    display_tracking_info,
    get_dir_size,
)

from sohstationviewer.database.process_db import execute_db_dict, execute_db

from sohstationviewer.conf.constants import TM_FORMAT


class MainWindow(QtWidgets.QMainWindow, UIMainWindow):
    current_directory_changed = QtCore.Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setup_ui(self)
        """
        dir_names: list of absolute path of data set directory
        """
        self.dir_names: List[Path] = []
        """
        data_type: str - type of data set
        """
        self.data_type: str = 'Unknown'

        self.data_loader: DataLoader = DataLoader()
        self.data_loader.finished.connect(self.replot_loaded_data)

        """
        processing_log: [(message, type)] - record the progress of processing
        """
        self.processing_log: List[Tuple[str, LogType]] = []
        """
        forms_in_forms_menu: List of forms in forms_menu
        """
        self.forms_in_forms_menu: List[QtWidgets.QWidget] = []
        """
        req_soh_chans: list of State-Of-Health channels to read data
            from. For Reftek, the list of channels is fixed => may not need
        """
        self.req_soh_chans: List[str] = []
        """
        req_wf_chans: list of waveform channels to read data from.
             For Reftek, it is [int,] which is list of index of data stream
             to read data from.
        """
        self.req_wf_chans: List[Union[str, int]] = []
        """
        start_tm: start time of data to read
        end_tm: end time of data to read
        """
        self.start_tm: float = 0
        self.end_tm: float = 0
        """
        data_object: Object that keep data read from data set for plotting
        """
        self.data_object: Union[DataTypeModel, None] = None
        """
        min_gap: minimum minutes of gap length to be display on gap bar
        """
        self.min_gap: Union[float, None] = None
        """
        pref_soh_list_name: name of selected preferred channels list
        """
        self.pref_soh_list_name: str = ''
        """
        pref_soh_list: selected preferred channels
        """
        self.pref_soh_list: List[str] = []
        """
        pref_soh_list_data_type: data type of the preferred channels list
        """
        self.pref_soh_list_data_type: str = 'Unknown'
        # Options
        """
        date_format: format for date
        """
        self.date_format: str = 'YYYY-MM-DD'
        """
        mass_pos_volt_range_opt: option for map value/color of mass position
        """
        self.mass_pos_volt_range_opt: str = 'regular'
        """
        bit_weight_opt: option for bitweight
        """
        self.bit_weight_opt: str = ''  # currently only need one option
        self.get_channel_prefer()
        self.yyyy_mm_dd_action.triggered.emit()

        """
        waveform_dlg: widget to display waveform channels' plotting
        """
        self.waveform_dlg: SOHWidget = WaveformDialog(self)
        """
        tps_dlg: dialog to display time-power-squared of waveform channels
        """
        self.tps_dlg: TimePowerSquaredDialog = TimePowerSquaredDialog(self)
        """
        help_browser: Display help documents with searching feature.
        """
        self.help_browser: HelpBrowser = HelpBrowser()
        """
        search_message_dialog: Display log, soh message with searching feature.
        """
        self.search_message_dialog: SearchMessageDialog = SearchMessageDialog()
        """
        gps_dialog: GPSDialog - widget to display plot of GPS points.
        """
        self.gps_dialog = GPSDialog(self)

        self.pull_current_directory_from_db()
        self.delete_old_temp_data_folder()

        self.has_problem: bool = False
        self.is_loading_data: bool = False
        self.is_plotting_soh: bool = False
        self.is_plotting_waveform: bool = False
        self.is_plotting_tps: bool = False
        self.is_stopping: bool = False

        self.background_black_radio_button.toggle()

    @QtCore.Slot()
    def open_data_type(self):
        """
        Open a dialog to add/edit data types in DB
        """
        win = DataTypeDialog(self)
        win.show()

    @QtCore.Slot()
    def open_param(self):
        """
        Open a dialog to add/edit parameters in DB
        """
        win = ParamDialog(self)
        win.show()

    @QtCore.Slot()
    def open_channel(self):
        """
        Open a dialog to add/edit channels in DB
        """
        win = ChannelDialog(self)
        win.show()

    @QtCore.Slot()
    def open_plot_type(self):
        """
        Open a dialog to view plot types and their description
        """
        win = PlotTypeDialog(self)
        win.show()

    @QtCore.Slot()
    def open_calendar(self):
        """
        Open calendar tool
        """
        calendar = CalendarDialog(self)
        calendar.show()

    @QtCore.Slot()
    def open_help_browser(self):
        """
        Open Help Dialog to view and search help documents
        """
        self.help_browser.show()
        self.help_browser.raise_()

    @QtCore.Slot()
    def open_channel_preferences(self):
        """
        Open a dialog to view, select, add, edit, scan for preferred channels
            list.
        """
        dir_names = [os.path.join(self.curr_dir_line_edit.text(), item.text())
                     for item in self.open_files_list.selectedItems()]
        if dir_names == []:
            msg = "No directories has been selected."
            QtWidgets.QMessageBox.warning(self, "Select directory", msg)
            return
        win = ChannelPreferDialog(self, dir_names)
        win.show()

    @QtCore.Slot()
    def all_wf_chans_clicked(self):
        if self.all_wf_chans_check_box.isChecked():
            self.mseed_wildcard_edit.clear()
            for cb in self.ds_check_boxes:
                cb.setChecked(False)

    @QtCore.Slot()
    def data_stream_clicked(self):
        ds_click = False
        for cb in self.ds_check_boxes:
            if cb.isChecked():
                ds_click = True
                break
        if ds_click:
            self.all_wf_chans_check_box.setChecked(False)

    @QtCore.Slot()
    def mseed_wildcard_changed(self):
        if self.mseed_wildcard_edit.text().strip() != "":
            self.all_wf_chans_check_box.setChecked(False)

    @QtCore.Slot()
    def all_soh_chans_clicked(self):
        """
        When "All SOH" is checked,
            + If checked, clear current IDs textbox
            + If unchecked, set current IDs textbox if there is a preferred
                channels list selected. If no list selected, re-check "All SOH"
        """
        if not self.all_soh_chans_check_box.isChecked():
            if self.pref_soh_list == []:
                self.all_soh_chans_check_box.setChecked(True)
            else:
                self.curr_pref_soh_list_name_txtbox.setText(
                    self.pref_soh_list_name)
        else:
            self.curr_pref_soh_list_name_txtbox.setText('')

    @QtCore.Slot()
    def set_date_format(self, display_format: str):
        """
        Sets the calendar format used by the QDateEdit text boxes.
        :param display_format: str - A valid display format to be used for date
            conversion.
        """
        display_format_to_qt_date_format_map = {
            'YYYY-MM-DD': 'yyyy-MM-dd',
            'YYYYMMMDD': 'yyyyMMMdd',
            'YYYY:DOY': 'yyyy:DOY'
        }
        qt_format = display_format_to_qt_date_format_map[display_format]
        self.time_to_date_edit.setDisplayFormat(qt_format)
        self.time_from_date_edit.setDisplayFormat(qt_format)
        self.date_format = display_format

    @QtCore.Slot()
    def open_files_list_item_double_clicked(self, item: FileListItem):
        """
        Handles the double-click event emitted when a user double-clicks on an
        item contained within openFilesList

        :param item: FileListItem - The item contained within openFilesList
            which was clicked.
        """
        print(f'Opening {item.text()}')
        self.read_selected_files()

    @QtCore.Slot()
    def change_current_directory(self):
        """
        Constructs a QFileDialog to select a new working directory from which
            the user can load data. The starting directory is taken from
            curr_dir_line_edit.
        """
        fd = QtWidgets.QFileDialog(self)
        fd.setFileMode(QtWidgets.QFileDialog.Directory)
        fd.setDirectory(self.curr_dir_line_edit.text())
        fd.exec_()
        try:
            new_path = fd.selectedFiles()[0]
            self.set_current_directory(new_path)
        except IndexError:
            pass

    def get_requested_wf_chans(self) -> List[Union[str, int]]:
        """
        Getting requested data stream for RT130 data or mseed wildcards for
            non-RT130 data
        :return req_wf_chans: list of data streams or list of mseed wildcards
        :rtype: List[str, int]
        """
        req_wf_chans = []
        if self.tps_check_box.isChecked() or self.raw_check_box.isChecked():
            if self.all_wf_chans_check_box.isChecked():
                req_mseed_wildcards = ['*']
                req_dss = ['*']      # all data stream
            else:
                req_dss = []
                req_mseed_wildcards = []
                for idx, ds_checkbox in enumerate(self.ds_check_boxes):
                    if ds_checkbox.isChecked():
                        req_dss.append(idx + 1)
                if self.mseed_wildcard_edit.text().strip() != "":
                    req_mseed_wildcards = self.mseed_wildcard_edit.text(
                        ).split(",")

            if self.data_type == 'RT130':
                req_wf_chans = req_dss
                if req_dss != ['*'] and req_mseed_wildcards != []:
                    msg = 'MSeed Wildcards will be ignored for RT130.'
                    self.processing_log.append((msg, LogType.WARNING))
            else:
                req_wf_chans = req_mseed_wildcards
                if req_mseed_wildcards != ['*'] and req_dss != []:
                    msg = ('Checked data streams will be ignored for '
                           'none-RT130 data type.')
                    self.processing_log.append((msg, LogType.WARNING))
        return req_wf_chans

    def get_requested_soh_chan(self):
        """
        Getting requested soh channels from preferred channel dialog
        :return list of requested soh channels
        :rtype: List[str]
        """
        if (not self.all_soh_chans_check_box.isChecked() and
                self.data_type != self.pref_soh_list_data_type):
            msg = (f"Data Type detected for the selected data set is "
                   f"{self.data_type} which is different from Channel "
                   f"Preferences {self.pref_soh_list_name}'s Data Type: "
                   f"{self.pref_soh_list_data_type}.\n\n"
                   f"SOHStationViewer will read all data available.\n\n"
                   f"Do you want to continue?")
            result = QtWidgets.QMessageBox.question(
                self, "Confirmation", msg,
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
            if result == QtWidgets.QMessageBox.No:
                return
            self.all_soh_chans_check_box.setChecked(True)
            self.curr_pref_soh_list_name_txtbox.setText('')
            return []
        return (self.pref_soh_list
                if not self.all_soh_chans_check_box.isChecked() else [])

    @QtCore.Slot()
    def read_selected_files(self):
        """
        Read data from selected files/directories, process and plot channels
            read from those according to current options set on the GUI
        """

        start_tm_str = self.time_from_date_edit.date().toString(
            QtCore.Qt.ISODate)
        end_tm_str = self.time_to_date_edit.date().toString(QtCore.Qt.ISODate)
        self.start_tm = datetime.strptime(start_tm_str, TM_FORMAT).timestamp()
        self.end_tm = datetime.strptime(end_tm_str, TM_FORMAT).timestamp()
        if self.end_tm <= self.start_tm:
            msg = "To Date must be greater than From Date."
            QtWidgets.QMessageBox.warning(self, "Wrong Date Given", msg)
            return
        self.info_list_widget.clear()
        is_working = (self.is_loading_data or self.is_plotting_soh or
                      self.is_plotting_waveform or self.is_plotting_tps)
        if is_working:
            msg = 'Already working'
            display_tracking_info(self.tracking_info_text_browser, msg, 'info')
            return
        self.has_problem = False

        try:
            check_chan_wildcards_format(self.mseed_wildcard_edit.text())
        except Exception as e:
            QtWidgets.QMessageBox.warning(self, "Incorrect Wildcard", str(e))
            return
        try:
            del self.data_object
            self.processing_log = []
            self.clear_actions_from_forms_menu()
        except AttributeError:
            pass

        self.req_soh_chans = (self.pref_soh_list
                              if not self.all_soh_chans_check_box.isChecked()
                              else [])

        self.dir_names = [
            Path(self.curr_dir_line_edit.text()).joinpath(item.text())
            for item in self.open_files_list.selectedItems()]
        if self.dir_names == []:
            msg = "No directories has been selected."
            QtWidgets.QMessageBox.warning(self, "Select directory", msg)
            return

        dir_size = sum(get_dir_size(str(dir))[0] for dir in self.dir_names)
        if dir_size > constants.BIG_FILE_SIZE:
            data_too_big_dialog = QMessageBox()
            data_too_big_dialog.setText('Chosen data set is very big. It '
                                        'might take a while to finish reading '
                                        'and plotting everything.')
            data_too_big_dialog.setInformativeText('Do you want to proceed?')
            data_too_big_dialog.setStandardButtons(QMessageBox.Yes |
                                                   QMessageBox.Abort)
            data_too_big_dialog.setDefaultButton(QMessageBox.Abort)
            data_too_big_dialog.setIcon(QMessageBox.Question)
            ret = data_too_big_dialog.exec_()
            if ret == QMessageBox.Abort:
                return

        self.data_type = detect_data_type(self.tracking_info_text_browser,
                                          self.dir_names)
        if self.data_type is None:
            return

        self.req_soh_chans = self.get_requested_soh_chan()
        self.req_wf_chans = self.get_requested_wf_chans()

        start_tm_str = self.time_from_date_edit.date().toString(
            QtCore.Qt.ISODate)
        end_tm_str = self.time_to_date_edit.date().toString(QtCore.Qt.ISODate)
        self.start_tm = datetime.strptime(start_tm_str, TM_FORMAT).timestamp()
        self.end_tm = datetime.strptime(end_tm_str, TM_FORMAT).timestamp()

        self.gps_dialog.clear_plot()
        self.gps_dialog.data_path = self.dir_names[0]
        self.gps_dialog.set_export_directory(str(self.dir_names[0]))
        self.gps_dialog.gps_points = []

        self.data_loader.init_loader(
            self.data_type,
            self.tracking_info_text_browser,
            self.dir_names,
            req_wf_chans=self.req_wf_chans,
            req_soh_chans=self.req_soh_chans,
            read_start=self.start_tm,
            read_end=self.end_tm,
            include_mp123=self.mass_pos_123zne_check_box.isChecked(),
            include_mp456=self.mass_pos_456uvw_check_box.isChecked()
        )

        self.data_loader.worker.finished.connect(self.data_loaded)
        self.data_loader.worker.stopped.connect(self.problem_happened)
        self.data_loader.worker.failed.connect(self.problem_happened)
        self.data_loader.thread.finished.connect(self.reset_flags)
        self.data_loader.connect_worker_signals()

        self.is_loading_data = True
        self.data_loader.load_data()

    def stop_load_data(self):
        # TODO: find a way to stop the data loader without a long wait.
        """
        Request the data loader thread to stop. The thread will stop at the
        earliest possible point, meaning that the wait is variable and can be
        very long.
        """
        if self.data_loader.running:
            self.is_loading_data = False
            self.data_loader.thread.requestInterruption()
            display_tracking_info(self.tracking_info_text_browser,
                                  'Stopping data loading...',
                                  LogType.INFO)

    @QtCore.Slot()
    def problem_happened(self):
        self.has_problem = True

    @QtCore.Slot()
    def stop(self):
        is_working = (self.is_loading_data or self.is_plotting_soh or
                      self.is_plotting_waveform or self.is_plotting_tps)
        if is_working:
            if self.is_stopping:
                msg = 'Already stopping'
                display_tracking_info(self.tracking_info_text_browser, msg,
                                      LogType.INFO)
                return
            self.is_stopping = True

        if self.is_loading_data:
            self.stop_load_data()
        if self.is_plotting_soh:
            display_tracking_info(self.tracking_info_text_browser,
                                  'Stopping SOH plot...')
        if self.is_plotting_waveform:
            display_tracking_info(self.waveform_dlg.info_text_browser,
                                  'Stopping waveform plot...')
            waveform_widget = self.waveform_dlg.plotting_widget
            waveform_widget.request_stop()

            self.waveform_dlg.plotting_widget.request_stop()
        if self.is_plotting_tps:
            display_tracking_info(self.tps_dlg.info_text_browser,
                                  'Stopping TPS plot...')
            tps_widget = self.tps_dlg.plotting_widget
            tps_widget.request_stop()

    def check_if_all_stopped(self):
        """
        Check if everything has been stopped. If true, reset the is_stopping
        flag.
        """
        not_all_stopped = (self.is_loading_data or self.is_plotting_soh or
                           self.is_plotting_waveform or self.is_plotting_tps)
        if not not_all_stopped:
            self.is_stopping = False

    @QtCore.Slot()
    def data_loaded(self, data_obj: DataTypeModel):
        """
        Process the loaded data.
        :param data_obj: the data object that contains the loaded data.
        """
        self.is_loading_data = False
        self.data_object = data_obj
        try:
            self.gps_dialog.gps_points = extract_gps_data(data_obj)
        except ValueError as e:
            missing_gps_chans = e.args[0].split(': ')[1].strip('.').split(', ')
            for missing_chan in missing_gps_chans:
                log_message = (f'GPS channel {missing_chan} is missing.',
                               LogType.WARNING
                               )
                self.processing_log.append(log_message)
            log_message = ('Cannot get GPS data.', LogType.WARNING)
            self.processing_log.append(log_message)

        data_set_info = extract_data_set_info(self.tracking_info_text_browser,
                                              data_obj, self.date_format)
        for info_name, info in data_set_info.items():
            if isinstance(info, str):
                info_string = f'{info_name}:\n\t{info}'
            else:
                formatted_info = '\n\t'.join(info)
                info_string = f'{info_name}:\n\t{formatted_info}'
            # Tab characters are rendered as ~20 spaces in QT, which is too
            # many considering the size of the display box. So, we replace all
            # tab characters with a number of spaces to make everything fits
            # better.
            info_string = info_string.replace('\t', ' ' * 4)
            QListWidgetItem(info_string, self.info_list_widget)

    @QtCore.Slot()
    def replot_loaded_data(self):
        """
        Plot using data from self.data_object with the current options set
            from GUI
        """
        if self.has_problem:
            return
        self.is_plotting_soh = True
        if self.detect_gap_check_box.isChecked():
            self.min_gap = self.gap_len_line_edit.text()
        else:
            self.min_gap = None
        do = self.data_object

        time_tick_total = 5  # TODO: let user choose max ticks to be displayed

        sel_key = do.selected_key

        soh_data = (deepcopy(do.soh_data[sel_key])
                    if sel_key in do.soh_data else {})
        soh_chans = list(soh_data.keys())

        mp_data = (deepcopy(do.mass_pos_data[sel_key])
                   if sel_key in do.mass_pos_data else {})

        gaps = (deepcopy(do.gaps[sel_key])
                if sel_key in do.gaps else {})
        try:
            check_masspos(mp_data, sel_key,
                          self.mass_pos_123zne_check_box.isChecked(),
                          self.mass_pos_456uvw_check_box.isChecked())
        except Exception as e:
            self.processing_log.append((str(e), LogType.WARNING))

        if len(self.req_wf_chans) != 0:
            try:
                wf_data = deepcopy(do.waveform_data[sel_key]['readData'])
            except KeyError:
                wf_data = {}
        else:
            wf_data = {}
        try:
            self.plotting_widget.plot_channels(
                self.start_tm, self.end_tm, sel_key,
                do.data_time[sel_key], soh_chans, time_tick_total,
                soh_data, mp_data, gaps)
        except Exception:
            fmt = traceback.format_exc()
            msg = f"Can't plot SOH data due to error: {str(fmt)}"
            display_tracking_info(self.tracking_info_text_browser, msg,
                                  LogType.ERROR)
            self.reset_flags()
        finally:
            self.is_plotting_soh = False

        peer_plotting_widgets = [self.plotting_widget]

        if self.tps_check_box.isChecked():
            self.is_plotting_tps = True
            peer_plotting_widgets.append(self.tps_dlg.plotting_widget)
            self.tps_dlg.set_data(
                self.data_type, ','.join([str(d) for d in self.dir_names]))
            self.tps_dlg.show()
            # The waveform and TPS plots is being stopped at the same time, so
            # we can't simply reset all flags. Instead, we use an intermediate
            # method that check whether all plots have been stopped before
            # resetting the is_stopping flag.
            tps_widget = self.tps_dlg.plotting_widget
            tps_widget.stopped.connect(self.reset_is_plotting_tps)
            tps_widget.stopped.connect(self.check_if_all_stopped)
            self.tps_dlg.plotting_widget.plot_channels(
                self.start_tm, self.end_tm, sel_key,
                do.data_time[sel_key],
                wf_data)
            self.add_action_to_forms_menu('TPS Plot', self.tps_dlg)
        else:
            self.tps_dlg.hide()

        if self.raw_check_box.isChecked():
            self.is_plotting_waveform = True
            # waveformPlot
            peer_plotting_widgets.append(self.waveform_dlg.plotting_widget)
            self.waveform_dlg.set_data(
                self.data_type, ','.join([str(d) for d in self.dir_names]))
            self.waveform_dlg.show()
            waveform_widget = self.waveform_dlg.plotting_widget
            waveform_widget.stopped.connect(self.reset_is_plotting_waveform)
            waveform_widget.stopped.connect(self.check_if_all_stopped)
            self.waveform_dlg.plotting_widget.plot_channels(
                self.start_tm, self.end_tm, sel_key,
                do.data_time[sel_key], time_tick_total,
                wf_data, mp_data)
            self.add_action_to_forms_menu('Raw Data Plot', self.waveform_dlg)
        else:
            self.waveform_dlg.hide()

        self.plotting_widget.set_peer_plotting_widgets(peer_plotting_widgets)
        self.waveform_dlg.plotting_widget.set_peer_plotting_widgets(
            peer_plotting_widgets)
        self.tps_dlg.plotting_widget.set_peer_plotting_widgets(
            peer_plotting_widgets)

        processing_log = (self.processing_log +
                          do.processing_log +
                          self.plotting_widget.processing_log)
        self.search_message_dialog.setup_logview(
            sel_key, do.log_data, processing_log)
        self.search_message_dialog.show()
        self.add_action_to_forms_menu('Search Messages',
                                      self.search_message_dialog)

    @QtCore.Slot()
    def reset_flags(self):
        """
        Reset the activity flags. Intended to be called when something goes
        wrong to reset the state of the program. Each step of the program
        should instead reset their own flag when they are finished.
        """
        self.is_loading_data = False
        self.is_plotting_soh = False
        self.is_plotting_waveform = False
        self.is_plotting_tps = False
        self.is_stopping = False

    @QtCore.Slot()
    def reset_is_plotting_waveform(self):
        """
        Reset the is_plotting_waveform flag. Used because lambda does not allow
        assignment.
        """
        self.is_plotting_waveform = False

    @QtCore.Slot()
    def reset_is_plotting_tps(self):
        """
        Reset the is_plotting_tps flag. Used because lambda does not allow
        assignment.
        """
        self.is_plotting_tps = False

    def set_current_directory(self, path: str = '') -> None:
        """
        Update currentDirectory with path in DB table PersistentData.
        Set all directories under current directory to self.open_files_list.

        :param path: str - absolute path to current directory
        """
        # Remove entries when current directory changed
        self.open_files_list.clear()
        # Signal current_directory_changed, and gather list of files in new
        # current directory
        self.current_directory_changed.emit(path)
        execute_db(f'UPDATE PersistentData SET FieldValue="{path}" WHERE '
                   'FieldName="currentDirectory"')
        try:
            for dent in pathlib.Path(path).iterdir():
                if not dent.is_dir() or dent.name.startswith('.'):
                    continue

                self.open_files_list.addItem(FileListItem(dent))
        except FileNotFoundError:
            self.set_current_directory()

    def get_channel_prefer(self):
        """
        Read the current preferred channel list from database to set
            pref_soh_list_name, pref_soh_list, self.pref_soh_list_data_type
        """
        self.pref_soh_list_name = ''
        self.pref_soh_list = []
        self.data_type = 'Unknown'
        rows = execute_db_dict('SELECT name, IDs, dataType FROM ChannelPrefer '
                               'WHERE current=1')
        if len(rows) > 0:
            self.pref_soh_list_name = rows[0]['name']
            self.pref_soh_list = [t.strip() for t in rows[0]['IDs'].split(',')]
            self.pref_soh_list_data_type = rows[0]['dataType']

    def resizeEvent(self, event):
        """
        OVERRIDE Qt method.
        When main_window is resized, its plotting_widget need to initialize
            its size to fit the viewport.

        :param event: QResizeEvent - resize event
        """
        self.plotting_widget.init_size()

    def pull_current_directory_from_db(self):
        """
        Set current directory with info saved in DB
        """
        rows = execute_db_dict(
            'SELECT FieldName, FieldValue FROM PersistentData '
            'WHERE FieldName="currentDirectory"')
        if len(rows) > 0 and rows[0]['FieldValue']:
            self.set_current_directory(rows[0]['FieldValue'])

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        """
        Cleans up when the user exits the program. Currently only clean up
        running data loaders.

        :param event: parameter of method being overridden
        """
        display_tracking_info(self.tracking_info_text_browser,
                              'Cleaning up...',
                              'info')
        if self.data_loader.running:
            self.data_loader.thread.requestInterruption()
            self.data_loader.thread.quit()
            self.data_loader.thread.wait()

        # close all remaining windows
        for window in QtWidgets.QApplication.topLevelWidgets():
            window.close()

    def delete_old_temp_data_folder(self) -> None:
        """
        Delete temp_data_folder which is used for keeping memmap files in case
            SOHView wasn't closed correctly.
        """
        rows = execute_db(
            'SELECT FieldValue FROM PersistentData '
            'WHERE FieldName="tempDataDirectory"')
        temp_data_folder = rows[0][0]
        try:
            shutil.rmtree(temp_data_folder)
            execute_db(
                f'UPDATE PersistentData SET FieldValue="{None}" WHERE'
                f' FieldName="tempDataDirectory"'
            )
        except (FileNotFoundError, TypeError):
            pass

    def add_action_to_forms_menu(
            self, form_name: str, form: QtWidgets.QWidget) -> None:
        """
        Creating and adding an action to forms_menu, connect to function
            raise_form() to help opening and raising the form on top of
            others when triggering the action.
        Adding the form into forms_in_forms_menu to handle closing it when
            loading new data set.

        :param form_name: name of form that will be shown in forms_menu
        :type form_name: str
        :param form: QtWidget
        :type form: QtWidgets.QWidget
        """
        if not isinstance(form, QtWidgets.QWidget):
            print(f"DEVELOPING ERROR: Type of form must be 'QWidget' instead "
                  f"of '{form.__class__.__name__}'")
        if form not in self.forms_in_forms_menu:
            action = QtWidgets.QAction(form_name, self)
            self.forms_menu.addAction(action)
            action.triggered.connect(lambda: self.raise_form(form))
            self.forms_in_forms_menu.append(form)
            self.forms_menu.removeAction(self.no_forms_action)

    def clear_actions_from_forms_menu(self) -> None:
        """
        Remove all actions from forms_menu.
        Close all forms related to forms_menu.
        Add no_forms_action to let user know that there are no forms available.
        """
        self.forms_menu.clear()     # remove all actions
        for i in range(len(self.forms_in_forms_menu) - 1, -1, -1):
            form = self.forms_in_forms_menu.pop(i)
            form.close()
        self.forms_menu.addAction(self.no_forms_action)

    @QtCore.Slot(QtWidgets.QWidget)
    def raise_form(self, form: QtWidgets.QWidget) -> None:
        form.show()                 # in case form has been closed
        form.raise_()               # to raise form on top of others

    @QtCore.Slot()
    def open_gps_plotter(self):
        self.gps_dialog.show()
        self.gps_dialog.raise_()

    @QtCore.Slot()
    def set_plots_color(self, checked, color_mode):
        if not checked:
            return
        self.gps_dialog.set_colors(color_mode)
