import os
from pathlib import Path
from tempfile import TemporaryDirectory

from unittest import TestCase
from unittest.mock import patch
from contextlib import redirect_stdout
import io

from sohstationviewer.model.handling_data import (
    read_mseed_or_text
)
from sohstationviewer.view.util.enums import LogType


TEST_DATA_DIR = Path(__file__).resolve().parent.parent.joinpath('test_data')
unimplemented_ascii_blockette_file = TEST_DATA_DIR.joinpath(
    'Q330_unimplemented_ascii_block/XX-3203_4-20221222190255')

log_text_file = TEST_DATA_DIR.joinpath(
    'Pegasus-sample/Pegasus_SVC4/logs/2020/XX/KC01/XX.KC01...D.2020.129')


def mock_track_info(text: str, type: LogType) -> None:
    print(text)


class TestReadMseedText(TestCase):
    def setUp(self) -> None:
        self.tmp_dir_obj: TemporaryDirectory = TemporaryDirectory()
        self.tmp_dir = self.tmp_dir_obj.name
        self.soh_streams = {}
        self.mass_pos_streams = {}
        self.waveform_data = {}
        self.data_time = {}
        self.log_data = {'TEXT': []}
        self.track_info = mock_track_info
        self.expected_soh_chan_ids = sorted([
            'VEP', 'VEC', 'VEA', 'VKI', 'VCO', 'VPB', 'LCQ', 'LCE', ])
        self.expected_mass_pos_chan_ids = sorted([
            'VMU', 'VMW', 'VMV'])
        self.expected_wf_chan_ids = sorted([
            'BH1', 'BH2', 'BH3', 'BH4', 'BH5', 'BH6',
            'EL1', 'EL2', 'EL4', 'EL5', 'EL6', 'ELZ',
            'LL1', 'LL2', 'LL4', 'LL5', 'LL6', 'LLZ',
            ])
        self.expected_spr_gt_1 = sorted([
            'BS1', 'BS2', 'BS3', 'BS4', 'BS5', 'BS6',
            'ES1', 'ES2', 'ES3', 'ES4', 'ES5', 'ES6',
            'LS1', 'LS2', 'LS3', 'LS4', 'LS5', 'LS6',
            'SS1', 'SS2', 'SS3', 'SS4', 'SS5', 'SS6'
        ])
        self.expected_log_chan_ids = sorted(['ACE', 'LOG', 'OCF'])
        self.expected_chan_ids = sorted(self.expected_soh_chan_ids +
                                        self.expected_mass_pos_chan_ids +
                                        self.expected_wf_chan_ids +
                                        self.expected_log_chan_ids +
                                        self.expected_spr_gt_1)

        self.expected_data_time = [1671729273.000004, 1671735657.9099982]

    @patch('sohstationviewer.model.mseed.blockettes_reader.'
           'NOT_READ_BLOCKETTES', {1000, 1001})
    def test_unimplemented_ascii_block_in_not_read_blockettes(self):
        """
        including testing ASCII text got from data
        """
        path2file = unimplemented_ascii_blockette_file
        f = io.StringIO()
        with redirect_stdout(f):
            ret = read_mseed_or_text(
                path2file, self.tmp_dir, None, None,
                self.soh_streams, self.mass_pos_streams, self.waveform_data,
                self.data_time, self.log_data, [], ['*'], True, True,
                self.track_info)
        output = f.getvalue()
        self.assertEqual(output.strip(), '')
        self.assertListEqual(ret['chan_ids'], self.expected_chan_ids)
        self.assertEqual(list(self.soh_streams.keys()), ['3203'])
        self.assertEqual(sorted(list(self.soh_streams['3203'].keys())),
                         self.expected_soh_chan_ids)
        self.assertEqual(list(self.mass_pos_streams.keys()), ['3203'])
        self.assertEqual(sorted(list(self.mass_pos_streams['3203'].keys())),
                         self.expected_mass_pos_chan_ids)
        self.assertEqual(list(self.waveform_data.keys()), ['3203'])
        self.assertEqual(
            sorted(list(self.waveform_data['3203']['readData'].keys())),
            self.expected_wf_chan_ids)
        self.assertEqual(list(self.data_time.keys()), ['3203'])
        self.assertListEqual(self.data_time['3203'],
                             self.expected_data_time)
        self.assertEqual(list(self.log_data.keys()), ['TEXT', '3203'])
        self.assertListEqual(sorted(list(self.log_data['3203'].keys())),
                             self.expected_log_chan_ids)
        self.assertEqual(
            self.log_data['3203']['LOG'][0].split('\n')[2],
            "STATE OF HEALTH: From:2022-12-22T18:54:33.200000Z  "
            "To:2022-12-22T18:54:33.200000Z")
        self.assertEqual(
            self.log_data['3203']['LOG'][0].split('\n')[3],
            "{800}[2022-12-22 18:54:33] PWRMGR: req=0 medtmr=9 ctr=1846 "
            "uvdc=12.977 ud=0.004 pd=0.005 pvdc=13.325 chg=0.027 low=0 "
            "trf=0 ms=0 ac=180 PM=GG\r")
        self.assertEqual(len(os.listdir(self.tmp_dir)), 36)

    @patch('sohstationviewer.model.mseed.blockettes_reader.'
           'NOT_READ_BLOCKETTES', {})
    def test_unimplemented_ascii_block_not_in_not_read_blockettes(self):
        path2file = unimplemented_ascii_blockette_file
        f = io.StringIO()
        with redirect_stdout(f):
            ret = read_mseed_or_text(
                path2file, self.tmp_dir, None, None,
                self.soh_streams, self.mass_pos_streams, self.waveform_data,
                self.data_time, self.log_data, [], ['*'], True, True,
                self.track_info)
        output = f.getvalue()
        self.assertEqual(
            output.strip(),
            f"{path2file.name}: 3203 - ACE: "
            f"Function to read blockette 1001 isn't implemented yet.\n"
            f"{path2file.name}: 3203 - OCF: "
            f"Function to read blockette 1001 isn't implemented yet.")

        self.assertListEqual(ret['chan_ids'], self.expected_chan_ids)
        self.assertEqual(list(self.soh_streams.keys()), ['3203'])
        self.assertEqual(sorted(list(self.soh_streams['3203'].keys())),
                         self.expected_soh_chan_ids)
        self.assertEqual(list(self.waveform_data.keys()), ['3203'])
        self.assertEqual(
            sorted(list(self.waveform_data['3203']['readData'].keys())),
            self.expected_wf_chan_ids)
        self.assertEqual(list(self.data_time.keys()), ['3203'])
        self.assertListEqual(self.data_time['3203'],
                             self.expected_data_time)
        self.assertEqual(list(self.log_data.keys()), ['TEXT', '3203'])
        self.assertListEqual(sorted(list(self.log_data['3203'].keys())),
                             self.expected_log_chan_ids)
        self.assertEqual(len(os.listdir(self.tmp_dir)), 36)

    def test_set_time_range(self):
        path2file = unimplemented_ascii_blockette_file
        expected_soh_chan_ids = ['LCE', 'LCQ']
        expected_mass_pos_chan_ids = ['VMU', 'VMV', 'VMW']
        expected_wf_chan_ids = ['LL1', 'LL2', 'LL4', 'LL5', 'LL6', 'LLZ']

        expected_chan_ids = sorted(expected_soh_chan_ids +
                                   expected_mass_pos_chan_ids +
                                   expected_wf_chan_ids)
        ret = read_mseed_or_text(
            path2file, self.tmp_dir, 1671730000, 1671730001,
            self.soh_streams, self.mass_pos_streams, self.waveform_data,
            self.data_time, self.log_data, [], ['*'], True, True,
            self.track_info)

        self.assertListEqual(ret['chan_ids'], expected_chan_ids)
        self.assertEqual(list(self.soh_streams.keys()), ['3203'])
        self.assertEqual(sorted(list(self.soh_streams['3203'].keys())),
                         expected_soh_chan_ids)
        self.assertEqual(list(self.waveform_data.keys()), ['3203'])
        self.assertEqual(
            sorted(list(self.waveform_data['3203']['readData'].keys())),
            expected_wf_chan_ids)
        self.assertEqual(list(self.data_time.keys()), ['3203'])
        self.assertListEqual(self.data_time['3203'],
                             [1671730000.000004, 1671730001.000004])
        self.assertEqual(list(self.log_data.keys()), ['TEXT'])

        self.assertEqual(len(os.listdir(self.tmp_dir)), 12)

    def test_text_file(self):
        path2file = log_text_file

        ret = read_mseed_or_text(
            path2file, self.tmp_dir, None, None,
            self.soh_streams, self.mass_pos_streams, self.waveform_data,
            self.data_time, self.log_data, [], ['*'], True, True,
            self.track_info)

        self.assertListEqual(ret['chan_ids'], [])
        self.assertEqual(list(self.soh_streams.keys()), [])
        self.assertEqual(list(self.waveform_data.keys()), [])
        self.assertEqual(list(self.data_time.keys()), [])
        self.assertEqual(list(self.log_data.keys()), ['TEXT'])

        self.assertEqual(len(os.listdir(self.tmp_dir)), 0)
